# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests
:Synopsis:       This package includes tests for the khoros library.
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  26 Mar 2020
"""